/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ColorSelectionFrame.h"
#include "ColorSelectionFrame_p.h"

#include "LineEditor.h"
#include "ColorGrabObject.h"
#include "DragMonitor.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "IconSize.h"
#include "InformationDialog.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDrag>
#include <QLayout>
#include <QMimeData>
#include <QPainter>
#include <QToolButton>

namespace Private
{

    //___________________________________________________________________________________
    ColorWidget::ColorWidget( QWidget* parent ):
        QPushButton( parent ),
        Counter( "ColorWidget" )
    {
        Debug::Throw( "ColorWidget::ColorWidget" );
        setIcon( IconEngine::get( IconNames::SelectColor ) );
        setText( tr( "Pick a New Color" ) );

        ColorGrabObject* object = new ColorGrabObject( this );
        connect( object, SIGNAL(colorSelected(QColor)), this, SIGNAL(colorSelected(QColor)) );
    }

    //___________________________________________________________________________________
    ColorDisplay::ColorDisplay( QWidget* parent ):
        QFrame( parent ),
        Counter( "ColorDisplay" )
    {
        Debug::Throw( "ColorDisplay::ColorDisplay" );
        setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
        setMinimumSize( IconSize::get( IconSize::Huge ) );

        DragMonitor* dragMonitor = new DragMonitor( this );
        connect( dragMonitor, SIGNAL(dragStarted(QPoint)), SLOT(_startDrag(QPoint)) );

    }

    //___________________________________________________________________________________
    void ColorDisplay::setColor( const QColor& color )
    { color_ = color; }

    //_______________________________________________
    void ColorDisplay::_startDrag( QPoint dragOrigin )
    {

        Debug::Throw( "ColorDisplay::_startDrag.\n" );

        // start drag
        QDrag *drag = new QDrag(this);
        QMimeData *mimeData = new QMimeData;
        mimeData->setData( "text/plain", qPrintable( color_.name() ) );
        drag->setMimeData( mimeData );

        // create drag pixmap
        QPixmap pixmap( size() );
        pixmap.fill( Qt::transparent );
        QPainter painter( &pixmap );
        _paint( &painter );

        drag->setPixmap( pixmap );
        drag->setHotSpot( dragOrigin-rect().topLeft() );

        drag->start();

    }

    //___________________________________________________________________________________
    void ColorDisplay::paintEvent( QPaintEvent* event )
    {
        if( color_.isValid() )
        {
            QPainter painter( this );
            painter.setClipRegion( event->region() );
            _paint( &painter );

            painter.setPen( Qt::NoPen );
            painter.end();

        }

        QFrame::paintEvent( event );

    }

    //___________________________________________________________________________________
    void ColorDisplay::_paint( QPainter* painter )
    {

        painter->setPen( Qt::NoPen );

        const int frameWidth( this->frameWidth() );
        const QRect rect( frameRect().adjusted( frameWidth, frameWidth, -frameWidth, -frameWidth ) );
        if( color_.alpha() == 255 )
        {

            // opaque color
            painter->setBrush( color_ );
            painter->drawRect( rect  );

        } else {

            // right side is transparent
            const qreal halfWidth = qreal( rect.width() )/2;

            const QRectF leftRect( rect.topLeft(), QSizeF( halfWidth, rect.height() ) );
            const QRectF rightRect( QPointF( halfWidth, 0 ) + rect.topLeft(),  QSizeF( halfWidth, rect.height() ) );
            painter->setBrush( color_ );
            painter->drawRect( rightRect );

            // left side is opaque
            painter->setBrush( QColor( color_.rgb() ) );
            painter->drawRect( leftRect );

        }

    }
}


//_________________________________________________________
ColorSelectionFrame::ColorSelectionFrame( QWidget* parent ):
    QWidget( parent ),
    Counter("ColorSelectionFrame" ),
    locked_( false )
{
    Debug::Throw( "ColorSelectionFrame::ColorSelectionFrame.\n" );

    _installActions();

    auto mainLayout = new QVBoxLayout;
    setLayout( mainLayout );

    // color widget
    mainLayout->addWidget( colorDisplay_ = new Private::ColorDisplay( this ), 0 );

    // display grid
    auto gridLayout = new GridLayout;
    gridLayout->setMargin(0);
    gridLayout->setMaxCount(4);
    gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
    gridLayout->setColumnAlignment( 2, Qt::AlignRight|Qt::AlignVCenter );
    mainLayout->addLayout( gridLayout );

    gridLayout->addWidget( new QLabel( "Name:", this ) );

    auto hLayout = new QHBoxLayout;
    hLayout->setMargin(0);
    gridLayout->addLayout( hLayout, 0, 1, 1, 3 );

    // editor
    hLayout->addWidget( editor_ = new LineEditor( this ) );
    editor_->setAlignment( Qt::AlignCenter );
    connect( editor_, SIGNAL(textEdited(QString)), SLOT(_fromEditor(QString)) );

    // save button
    auto button = new QToolButton( this );
    button->setDefaultAction( saveAction_ );
    button->setAutoRaise( true );

    button->setIcon( IconEngine::get( IconNames::Save ) );
    button->setToolTip( tr( "Save current color to list" ) );
    hLayout->addWidget( button, 0 );

    // spinboxes
    SpinBox *spinbox;
    gridLayout->addWidget( new QLabel( tr( "Red:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Red ) );
    spinboxes_.insert( Red,  spinbox );

    gridLayout->addWidget( new QLabel( tr( "Hue:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Hue ) );
    spinboxes_.insert( Hue, spinbox );

    gridLayout->addWidget( new QLabel( tr( "Green:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Green ) );
    spinboxes_.insert( Green, spinbox );

    gridLayout->addWidget( new QLabel( tr( "Saturation:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Saturation ) );
    spinboxes_.insert( Saturation,  spinbox );

    gridLayout->addWidget( new QLabel( tr( "Blue:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Blue ) );
    spinboxes_.insert( Blue, spinbox );

    gridLayout->addWidget( new QLabel( tr( "Value:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Value ) );
    spinboxes_.insert( Value, spinbox );

    gridLayout->addWidget( new QLabel( tr( "Alpha:" ), this ) );
    gridLayout->addWidget( spinbox = new SpinBox( this, Alpha ) );
    spinboxes_.insert( Alpha,  spinbox );

    gridLayout->setColumnStretch( 1, 1 );
    gridLayout->setColumnStretch( 3, 1 );

    for( auto&& iter = spinboxes_.begin(); iter!=spinboxes_.end(); ++iter )
    {
        iter.value()->setMinimum(0);
        iter.value()->setMaximum(255);
        iter.value()->setValue(0);
    }

    spinboxes_[Alpha]->setValue(255);

    // color selection widget
    mainLayout->addWidget( colorWidget_ = new Private::ColorWidget( this ), 0, Qt::AlignCenter );
    connect( colorWidget_, SIGNAL(colorSelected(QColor)), SLOT(setColor(QColor)) );

    // connections
    connect( spinboxes_[Hue], SIGNAL(valueChanged(int)), SLOT(_fromHsv()) );
    connect( spinboxes_[Saturation], SIGNAL(valueChanged(int)), SLOT(_fromHsv()) );
    connect( spinboxes_[Value], SIGNAL(valueChanged(int)), SLOT(_fromHsv()) );

    connect( spinboxes_[Red], SIGNAL(valueChanged(int)), SLOT(_fromRgb()) );
    connect( spinboxes_[Green], SIGNAL(valueChanged(int)), SLOT(_fromRgb()) );
    connect( spinboxes_[Blue], SIGNAL(valueChanged(int)), SLOT(_fromRgb()) );
    connect( spinboxes_[Alpha], SIGNAL(valueChanged(int)), SLOT(_fromRgb()) );

    // configuration update
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );

    return;
}

//_________________________________________________________
QAbstractButton& ColorSelectionFrame::colorWidget() const
{ return *colorWidget_; }

//_________________________________________________________
void ColorSelectionFrame::setColor( QColor color )
{
    // check color
    if( !color.isValid() )
    {
        InformationDialog( this, tr( "Selected color is invalid" ) ).exec();
        return;
    }

    // update color
    color_ = color;
    Debug::Throw() << "ColorSelectionFrame::setColor - color set" << endl;

    spinboxes_[Red]->setValue( color.red() );
    spinboxes_[Green]->setValue( color.green() );
    spinboxes_[Blue]->setValue( color.blue() );

    int hue( 0 ), saturation( 0 ), value( 0 );
    color.getHsv( &hue, &saturation, &value );
    spinboxes_[Hue]->setValue( hue );
    spinboxes_[Saturation]->setValue( saturation );
    spinboxes_[Value]->setValue( value );

    spinboxes_[Alpha]->setValue( color.alpha() );

    // update label
    _updateLabel();

}


//_________________________________________________________
void ColorSelectionFrame::_fromEditor( const QString& text )
{
    Debug::Throw( "ColorSelectionFrame::_fromEditor.\n" );

    if( locked_ ) return;
    locked_ = true;

    // retrieve RGB indexes
    QColor color( text );
    if( !color.isValid() )
    {
        locked_ = false;
        return;
    }

    int hue( 0 ), saturation( 0 ), value( 0 );
    color.getHsv( &hue, &saturation, &value );

    spinboxes_[Hue]->setValue(hue);
    spinboxes_[Saturation]->setValue(saturation);
    spinboxes_[Value]->setValue(value);

    int red( color.red() );
    int green( color.green() );
    int blue( color.blue() );

    spinboxes_[Red]->setValue(red);
    spinboxes_[Green]->setValue(green);
    spinboxes_[Blue]->setValue(blue);

    // update color and display
    color_ = color;
    colorDisplay_->setColor( color_ );
    colorDisplay_->update();

    locked_ = false;
}

//_________________________________________________________
void ColorSelectionFrame::_fromRgb()
{
    Debug::Throw( "ColorSelectionFrame::_fromRgb.\n" );

    if( locked_ ) return;
    locked_ = true;

    // retrieve RGB indexes
    int red( spinboxes_[Red]->value() );
    int green( spinboxes_[Green]->value() );
    int blue( spinboxes_[Blue]->value() );
    int alpha( spinboxes_[Alpha]->value() );

    QColor color;
    color.setRgba( qRgba( red, green, blue, alpha ) );
    if( !color.isValid() )
    {
        locked_ = false;
        return;
    }

    int hue( 0 ), saturation( 0 ), value( 0 );
    color.getHsv( &hue, &saturation, &value );

    spinboxes_[Hue]->setValue(hue);
    spinboxes_[Saturation]->setValue(saturation);
    spinboxes_[Value]->setValue(value);

    // update color
    color_ = color;

    // update label
    _updateLabel();
    locked_ = false;

}

//_________________________________________________________
void ColorSelectionFrame::_fromHsv()
{
    Debug::Throw( "ColorSelectionFrame::_fromHsv.\n" );

    if( locked_ ) return;
    locked_ = true;

    // retrieve RGB indexes
    int hue( spinboxes_[Hue]->value() );
    int saturation( spinboxes_[Saturation]->value() );
    int value( spinboxes_[Value]->value() );

    QColor color;
    color.setHsv( hue, saturation, value );
    if( !color.isValid() ) {
        locked_ = false;
        return;
    }

    int red( 0 ), green( 0 ), blue( 0 );
    color.getRgb( &red, &green, &blue );

    spinboxes_[Red]->setValue(red);
    spinboxes_[Green]->setValue(green);
    spinboxes_[Blue]->setValue(blue);

    // update color
    color_ = color;

    // update label
    _updateLabel();
    locked_ = false;
}

//_________________________________________________________
void ColorSelectionFrame::_updateLabel()
{

    Debug::Throw( "ColorSelectionFrame::_updateLabel.\n" );

    // update color widget
    colorDisplay_->setColor( color_ );
    colorDisplay_->update();

    // update label text
    editor_->setText( color_.name() );

}

//_________________________________________________________
void ColorSelectionFrame::_toggleAutoHide( bool value )
{
    Debug::Throw( "ColorSelectionFrame::_toggleAutoHide.\n" );
    XmlOptions::get().set<bool>( "AUTO_HIDE", value );
    if( value )
    {

        connect( colorWidget_, SIGNAL(clicked()), window(), SLOT(hide()));
        connect( colorWidget_, SIGNAL(colorSelected(QColor)), window(), SLOT(show()));

    } else {

        disconnect( colorWidget_, SIGNAL(clicked()), window(), SLOT(hide()));
        disconnect( colorWidget_, SIGNAL(colorSelected(QColor)), window(), SLOT(show()));

    }
}

//_________________________________________________________
void ColorSelectionFrame::_updateConfiguration()
{
    Debug::Throw( "ColorSelectionFrame::_updateConfiguration.\n" );
    autoHideAction_->setChecked( XmlOptions::get().get<bool>( "AUTO_HIDE" ) );
}

//_________________________________________________________________________
void ColorSelectionFrame::_installActions()
{
    Debug::Throw( "ColorSelectionFrame::_installActions.\n" );

    // auto hide
    addAction( autoHideAction_ = new QAction( tr( "Hide Application when Selecting Color" ), this ) );
    autoHideAction_->setCheckable( true );
    autoHideAction_->setChecked( false );
    connect(autoHideAction_, SIGNAL(toggled(bool)), SLOT(_toggleAutoHide(bool)));

    // save
    addAction( saveAction_ = new QAction( IconEngine::get( IconNames::Save ), tr( "Save" ), this ) );
    saveAction_->setToolTip( tr( "Add current color to list" ) );
    saveAction_->setShortcut( QKeySequence::Save );

}
